import numpy as np
from scipy.integrate import simpson
from scipy.special import erf

# TO-DO: Add nbins as another dimension to all quantities to remove looping!

class HOD:
    def __init__(
            self,
            mass=None,
            dndlnm=None,
            nz=1
        ):
        if mass is None or dndlnm is None:
            raise ValueError("Mass and halo mass function need to be specified!")

        # Set all given parameters.
        # With newaxis we make sure the HOD shape is (nz, nmass)
        self.mass = mass[np.newaxis, :]
        self.dndlnm = dndlnm
        self.nz = nz

    @property
    def compute_number_density_cen(self):
        """
        Total number density of galaxies with the given HOD, e.g. central and satellite galaxies
        This is an integral over the HOD and the halo mass function to remove the halo mass dependence.
        Nx = int ⟨Nx|M⟩ n(M) dM
        """
        integrand = self.compute_hod_cen * self.dndlnm / self.mass
        return simpson(integrand, self.mass)

    @property
    def compute_number_density_sat(self):
        """
        Total number density of galaxies with the given HOD, e.g. central and satellite galaxies
        This is an integral over the HOD and the halo mass function to remove the halo mass dependence.
        Nx = int ⟨Nx|M⟩ n(M) dM
        """
        integrand = self.compute_hod_sat * self.dndlnm / self.mass
        return simpson(integrand, self.mass)

    @property
    def compute_number_density(self):
        """
        Total number density of galaxies with the given HOD, e.g. central and satellite galaxies
        This is an integral over the HOD and the halo mass function to remove the halo mass dependence.
        Nx = int ⟨Nx|M⟩ n(M) dM
        """
        integrand = self.compute_hod * self.dndlnm / self.mass
        return simpson(integrand, self.mass)

    @property
    def compute_avg_halo_mass_cen(self):
        """
        The mean halo mass for the given population of galaxies, e.g. central and satellite galaxies
        M_mean = int ⟨Nx|M⟩ M n(M) dM
        """
        integrand = self.compute_hod_cen * self.dndlnm
        return simpson(integrand, self.mass)

    @property
    def compute_avg_halo_mass_sat(self):
        """
        The mean halo mass for the given population of galaxies, e.g. central and satellite galaxies
        M_mean = int ⟨Nx|M⟩ M n(M) dM
        """
        integrand = self.compute_hod_sat * self.dndlnm
        return simpson(integrand, self.mass)

    @property
    def compute_avg_halo_mass(self):
        """
        The mean halo mass for the given population of galaxies, e.g. central and satellite galaxies
        M_mean = int ⟨Nx|M⟩ M n(M) dM
        """
        integrand = self.compute_hod * self.dndlnm
        return simpson(integrand, self.mass)

    def compute_galaxy_linear_bias_cen(self, halo_bias):
        """
        Mean linear halo bias for the given population of galaxies.
        b_lin_x = int ⟨Nx|M⟩ b_h(M) n(M) dM
        """
        bg_integrand = self.compute_hod_cen * halo_bias * self.dndlnm / self.mass
        return simpson(bg_integrand, self.mass)

    def compute_galaxy_linear_bias_sat(self, halo_bias):
        """
        Mean linear halo bias for the given population of galaxies.
        b_lin_x = int ⟨Nx|M⟩ b_h(M) n(M) dM
        """
        bg_integrand = self.compute_hod_sat * halo_bias * self.dndlnm / self.mass
        return simpson(bg_integrand, self.mass)

    def compute_galaxy_linear_bias(self, halo_bias):
        """
        Mean linear halo bias for the given population of galaxies.
        b_lin_x = int ⟨Nx|M⟩ b_h(M) n(M) dM
        """
        bg_integrand = self.compute_hod * halo_bias * self.dndlnm / self.mass
        return simpson(bg_integrand, self.mass)

class Cacciato(HOD):
    """
    The conditional observable functions (COFs) tell us how many galaxies with the observed property O, exist in haloes of
    mass M: Φ(O|M).
    Integrating over the observable will give us the total number of galaxies in haloes of a given mass, the so-called
    Halo Occupation Distribution (HOD).
    The observable can be galaxy stellar mass or galaxy luminosity or possibly other properties of galaxies.
    Note that the general mathematical form of the COFs might not hold for other observables.
    COF is different for central and satellite galaxies. The total COF can be written as the sum of the two:
    Φ(O|M) = Φc(O|M) + Φs(O|M)
    The halo mass dependence comes in through pivot observable values denoted by *, e.g. O∗c, O∗s
    """
    def __init__(
            self,
            obs = np.array([np.logspace(9, 15, 100)]),
            log10_obs_norm_c = 9.95,
            log10_m_ch = 11.24,
            g1 = 3.18,
            g2 = 0.245,
            sigma_log10_O_c = 0.157,
            norm_s = 0.562,
            pivot = 12.0,
            alpha_s = -1.18,
            beta_s = 2,
            b0 = -1.17,
            b1 = 1.53,
            b2 = -0.217,
            A_cen = None,
            A_sat = None,
            **hod_kwargs
        ):

        # Call super init MUST BE DONE FIRST.
        super().__init__(**hod_kwargs)
        # Set all given parameters.
        # With newaxis we make sure the COF shape is (nz, nmass, nobs)
        self.obs = obs[:, np.newaxis]

        # centrals
        # all observable masses in units of log10(M_sun h^-2)
        self.M_char = 10.0**log10_m_ch  # M_char
        self.g_1 = g1  # gamma_1
        self.g_2 = g2  # gamma_2
        self.Obs_norm_c = 10.0**log10_obs_norm_c  # O_0, O_norm_c
        self.sigma_log10_O_c = sigma_log10_O_c  # sigma_log10_O_c
        # satellites
        self.norm_s = norm_s  # extra normalisation factor for satellites
        self.pivot = pivot  # pivot mass for the normalisation of the stellar mass function: ϕ∗s
        self.alpha_s = alpha_s  # goes into the conditional stellar mass function COF_sat(M*|M)
        self.beta_s = beta_s  # goes into the conditional stellar mass function COF_sat(M*|M)
        # log10[ϕ∗s(M)] = b0 + b1(log10 m_p)+ b2(log10 m_p)^2, m_p = M/pivot
        self.b0 = b0
        self.b1 = b1
        self.b2 = b2
        # Decorated HOD assembly bias parameters
        self.A_cen = A_cen
        self.A_sat = A_sat

    @property
    def COF_cen(self):
        """
        COF for Central galaxies.
        eq 17 of D23: 2210.03110:
        Φc(O|M) = 1/[√(2π) ln(10) σ_c O] exp[ -log(O/O∗c)^2/ (2 σ_c^2) ]
        Note Φc(O|M) is unitless.
        """
        mean_obs_c = self.cal_mean_obs_c[:, :, np.newaxis]  # O∗c
        COF_c = (1.0 / (np.sqrt(2.0 * np.pi) * np.log(10.0) * self.sigma_log10_O_c * self.obs) *
                 np.exp(-(np.log10(self.obs / mean_obs_c))**2 / (2.0 * self.sigma_log10_O_c**2)))
        return COF_c

    @property
    def COF_sat(self):
        """
        COF for satellite galaxies.
        eq 18 of D23: 2210.03110:
        Φs(O|M) = ϕ∗s/O∗s (O/O∗s)^α_s exp [−(O/O∗s)^2], O*s is O∗s(M) = 0.56 O∗c(M)
        Note Φs(O|M) is unitless.
        """
        obs_s_star = self.norm_s * self.cal_mean_obs_c[:, :, np.newaxis]
        obs_tilde = self.obs / obs_s_star
        phi_star_val = self.phi_star_s[:, :, np.newaxis]
        COF_s = (phi_star_val / obs_s_star) * (obs_tilde**self.alpha_s) * np.exp(-obs_tilde**self.beta_s)
        return COF_s

    @property
    def COF(self):
        return self.COF_cen + self.COF_sat

    def obs_func_cen(self, axis=-2):
        r"""
        The Observable function, this is Φs(O|M), Φc(O|M) integrated over the halo mass weighted
        with the Halo Mass Function (HMF) to give:  Φs(O),Φc(O)
        Φx(O) =int Φx(O|M) n(M) dM,
        dndlnm is basically n(M) x mass, it is the output of hmf
        The differential mass function in terms of natural log of m,
        len=len(m) [units \(h^3 Mpc^{-3}\)]
        dn(m)/ dln m eq1 of 1306.6721
        obs_func unit is h^3 Mpc^{-3} dex^-1
        """
        integrand = self.COF_cen * self.dndlnm[:, :, np.newaxis] / self.mass[:, :, np.newaxis]
        obs_function = simpson(integrand, self.mass[0, :], axis=axis)
        return obs_function

    def obs_func_sat(self, axis=-2):
        r"""
        The Observable function, this is Φs(O|M), Φc(O|M) integrated over the halo mass weighted
        with the Halo Mass Function (HMF) to give:  Φs(O),Φc(O)
        Φx(O) =int Φx(O|M) n(M) dM,
        dndlnm is basically n(M) x mass, it is the output of hmf
        The differential mass function in terms of natural log of m,
        len=len(m) [units \(h^3 Mpc^{-3}\)]
        dn(m)/ dln m eq1 of 1306.6721
        obs_func unit is h^3 Mpc^{-3} dex^-1
        """
        integrand = self.COF_sat * self.dndlnm[:, :, np.newaxis] / self.mass[:, :, np.newaxis]
        obs_function = simpson(integrand, self.mass[0, :], axis=axis)
        return obs_function

    def obs_func(self, axis=-2):
        r"""
        The Observable function, this is Φs(O|M), Φc(O|M) integrated over the halo mass weighted
        with the Halo Mass Function (HMF) to give:  Φs(O),Φc(O)
        Φx(O) =int Φx(O|M) n(M) dM,
        dndlnm is basically n(M) x mass, it is the output of hmf
        The differential mass function in terms of natural log of m,
        len=len(m) [units \(h^3 Mpc^{-3}\)]
        dn(m)/ dln m eq1 of 1306.6721
        obs_func unit is h^3 Mpc^{-3} dex^-1
        """
        integrand = self.COF * self.dndlnm[:, :, np.newaxis] / self.mass[:, :, np.newaxis]
        obs_function = simpson(integrand, self.mass[0, :], axis=axis)
        return obs_function

    @property
    def cal_mean_obs_c(self):
        """
        eqs 19 of D23: 2210.03110
        O∗c(M) = O_0 (M/M1)^γ1 / [1 + (M/M1)]^(γ1−γ2)
        To get the values for the satellite call this * hod_par.norm_s
        O∗s(M) = 0.56 O∗c(M)
        Here M1 is a characteristic mass scale, and O_0 is the normalization.
        used to be mor

        (observable can be galaxy luminosity or stellar mass)
        returns the observable given halo mass. Assumed to be a double power law with characteristic
        scale m_1, normalisation m_0 and slopes g_1 and g_2
        """
        mean_obs_c = (self.Obs_norm_c * (self.mass / self.M_char)**self.g_1 /
                      (1.0 + (self.mass / self.M_char))**(self.g_1 - self.g_2))
        return mean_obs_c

    @property
    def phi_star_s(self):
        """
        pivot COF used in eq 21 of D23: 2210.03110
        using a bias expansion around the pivot mass
        eq 22 of D23: 2210.03110
        log[ϕ∗s(M)] = b0 + b1(log m13) , m13 is logM_pivot, m13 = M/(hod_par.pivot M⊙ h−1)
        """
        logM_pivot = np.log10(self.mass) - self.pivot
        log_phi_s = self.b0 + self.b1 * logM_pivot + self.b2 * (logM_pivot**2.0)
        return 10.0**log_phi_s

    @property
    def compute_hod_cen(self):
        """
        The HOD is computed by integrating over the COFs
        eq 23 of D23: 2210.03110
        ⟨Nx|M⟩ =int_{O_low}^{O_high} Φx(O|M) dO
        """
        N_cen = simpson(self.COF_cen, self.obs)
        if self.A_cen is not None:
            delta_pop_c = self.A_cen * np.fmin(N_cen, 1.0 - N_cen)
            N_cen = N_cen + delta_pop_c
        return N_cen

    @property
    def compute_hod_sat(self):
        """
        The HOD is computed by integrating over the COFs
        eq 23 of D23: 2210.03110
        ⟨Nx|M⟩ =int_{O_low}^{O_high} Φx(O|M) dO
        """
        N_sat = simpson(self.COF_sat, self.obs)
        if self.A_sat is not None:
            delta_pop_s = self.A_sat * N_sat
            N_sat = N_sat + delta_pop_s
        return N_sat

    @property
    def compute_hod(self):
        return self.compute_hod_cen + self.compute_hod_sat

    @property
    def compute_stellar_fraction_cen(self):
        """
        The mean value of the observable for the given galaxy population for a given halo mass.
        O is weighted by the number of galaxies with the property O for each halo mass: Φx(O|M)
        f_star = int_{O_low}^{O_high} Φx(O|M) O dO
        """
        return simpson(self.COF_cen * self.obs, self.obs) / self.mass

    @property
    def compute_stellar_fraction_sat(self):
        """
        The mean value of the observable for the given galaxy population for a given halo mass.
        O is weighted by the number of galaxies with the property O for each halo mass: Φx(O|M)
        f_star = int_{O_low}^{O_high} Φx(O|M) O dO
        """
        return simpson(self.COF_sat * self.obs, self.obs) / self.mass

    @property
    def compute_stellar_fraction(self):
        """
        The mean value of the observable for the given galaxy population for a given halo mass.
        O is weighted by the number of galaxies with the property O for each halo mass: Φx(O|M)
        f_star = int_{O_low}^{O_high} Φx(O|M) O dO
        """
        return simpson(self.COF * self.obs, self.obs) / self.mass

class Simple(HOD):
    """
    Simple HOD model
    """
    def __init__(
            self,
            obs = None,
            log10_Mmin = 12.0,
            log10_Msat = 13.0,
            alpha = 1.0,
            A_cen = None,
            A_sat = None,
            **hod_kwargs
        ):
        # Call super init MUST BE DONE FIRST.
        super().__init__(**hod_kwargs)
        self.obs = obs
        self.Mmin = 10.0**log10_Mmin
        self.Msat = 10.0**log10_Msat
        self.alpha = alpha
        # Decorated HOD assembly bias parameters
        self.A_cen = A_cen
        self.A_sat = A_sat

    @property
    def compute_hod_cen(self):
        """
        Centrals
        """
        N_cen = np.heaviside(self.mass - self.Mmin, 1.0)
        if self.A_cen is not None:
            delta_pop_c = self.A_cen * np.fmin(N_cen, 1.0 - N_cen)
            N_cen = N_cen + delta_pop_c
        return np.tile(N_cen, (self.nz, 1))

    @property
    def compute_hod_sat(self):
        """
        Satellites
        """
        N_sat = self.compute_hod_cen * (self.mass / self.Msat)**self.alpha
        if self.A_sat is not None:
            delta_pop_s = self.A_sat * N_sat
            N_sat = N_sat + delta_pop_s
        return np.tile(N_sat, (self.nz, 1))

    @property
    def compute_hod(self):
        return self.compute_hod_cen + self.compute_hod_sat

class Zehavi(HOD):
    """
    HOD model from Zehavi et al. (2004; https://arxiv.org/abs/astro-ph/0703457)
    Same as Zheng model in the limit that sigma=0 and M0=0
    Mean number of central galaxies is only ever 0 or 1 in this HOD
    """
    def __init__(
            self,
            obs = None,
            log10_Mmin = 12.0,
            log10_Msat = 13.0,
            alpha = 1.0,
            A_cen = None,
            A_sat = None,
            **hod_kwargs
        ):
        # Call super init MUST BE DONE FIRST.
        super().__init__(**hod_kwargs)
        self.obs = obs
        self.Mmin = 10.0**log10_Mmin
        self.Msat = 10.0**log10_Msat
        self.alpha = alpha
        # Decorated HOD assembly bias parameters
        self.A_cen = A_cen
        self.A_sat = A_sat

    @property
    def compute_hod_cen(self):
        """
        Centrals
        """
        N_cen = np.heaviside(self.mass - self.Mmin, 1.0)
        if self.A_cen is not None:
            delta_pop_c = self.A_cen * np.fmin(N_cen, 1.0 - N_cen)
            N_cen = N_cen + delta_pop_c
        return np.tile(N_cen, (self.nz, 1))

    @property
    def compute_hod_sat(self):
        """
        Satellites
        """
        N_sat = (self.mass / self.Msat)**self.alpha
        if self.A_sat is not None:
            delta_pop_s = self.A_sat * N_sat
            N_sat = N_sat + delta_pop_s
        return np.tile(N_sat, (self.nz, 1))

    @property
    def compute_hod(self):
        return self.compute_hod_cen + self.compute_hod_sat

class Zheng(HOD):
    """
    Zheng et al. (2005; https://arxiv.org/abs/astro-ph/0408564) HOD model
    """
    def __init__(
            self,
            obs = None,
            log10_Mmin = 12.0,
            log10_M0 = 12.0,
            log10_M1 = 13.0,
            sigma = 0.15,
            alpha = 1.0,
            A_cen = None,
            A_sat = None,
            **hod_kwargs
        ):
        # Call super init MUST BE DONE FIRST.
        super().__init__(**hod_kwargs)
        self.obs = obs
        self.Mmin = 10.0**log10_Mmin
        self.M0 = 10.0**log10_M0
        self.M1 = 10.0**log10_M1
        self.sigma = sigma
        self.alpha = alpha
        # Decorated HOD assembly bias parameters
        self.A_cen = A_cen
        self.A_sat = A_sat

    @property
    def compute_hod_cen(self):
        """
        Centrals
        """
        if self.sigma == 0.0:
            N_cen = np.heaviside(self.mass - self.Mmin, 1.0)
        else:
            N_cen = 0.5 * (1.0 + erf(np.log10(self.mass / self.Mmin) / self.sigma))
        if self.A_cen is not None:
            delta_pop_c = self.A_cen * np.fmin(N_cen, 1.0 - N_cen)
            N_cen = N_cen + delta_pop_c
        return np.tile(N_cen, (self.nz, 1))

    @property
    def compute_hod_sat(self):
        """
        Satellites
        """
        N_sat = (np.heaviside(self.mass - self.M0, 1.0) * (self.mass - self.M0) / self.M1)**self.alpha
        if self.A_sat is not None:
            delta_pop_s = self.A_sat * N_sat
            N_sat = N_sat + delta_pop_s
        return np.tile(N_sat, (self.nz, 1))

    @property
    def compute_hod(self):
        return self.compute_hod_cen + self.compute_hod_sat

class Zhai(HOD):
    """
    HOD model from Zhai et al. (2017; https://arxiv.org/abs/1607.05383)
    """
    def __init__(
            self,
            obs = None,
            log10_Mmin = 13.68,
            log10_Msat = 14.87,
            log10_Mcut = 12.32,
            sigma = 0.82,
            alpha = 0.41,
            A_cen = None,
            A_sat = None,
            **hod_kwargs
        ):
        # Call super init MUST BE DONE FIRST.
        super().__init__(**hod_kwargs)
        self.obs = obs
        self.Mmin = 10.0**log10_Mmin
        self.Msat = 10.0**log10_Msat
        self.Mcut = 10.0**log10_Mcut
        self.sigma = sigma
        self.alpha = alpha
        # Decorated HOD assembly bias parameters
        self.A_cen = A_cen
        self.A_sat = A_sat

    @property
    def compute_hod_cen(self):
        """
        Centrals
        """
        if self.sigma == 0.0:
            N_cen = np.heaviside(self.mass - self.Mmin, 1.0)
        else:
            N_cen = 0.5 * (1.0 + erf(np.log10(self.mass / self.Mmin) / self.sigma))
        if self.A_cen is not None:
            delta_pop_c = self.A_cen * np.fmin(N_cen, 1.0 - N_cen)
            N_cen = N_cen + delta_pop_c
        return np.tile(N_cen, (self.nz, 1))

    @property
    def compute_hod_sat(self):
        """
        Satellites
        """
        # Paper has a Nc(M) multiplication, but I think the central condition covers this
        N_sat = ((self.mass / self.Msat)**self.alpha) * np.exp(-self.Mcut / self.mass)
        if self.A_sat is not None:
            delta_pop_s = self.A_sat * N_sat
            N_sat = N_sat + delta_pop_s
        return np.tile(N_sat, (self.nz, 1))

    @property
    def compute_hod(self):
        return self.compute_hod_cen + self.compute_hod_sat



import numpy as np
import pyfftlog
import scipy.interpolate
from cosmosis.datablock import option_section
from scipy.integrate import simpson

# These are the ones the user can use
TRANSFORM_WP = "wp"
TRANSFORM_DS = "ds"

DEFAULT_N_TRANSFORM = 8192
DEFAULT_K_MIN = 0.0001
DEFAULT_K_MAX = 5.0e6
DEFAULT_RP_MIN = 0.1
DEFAULT_RP_MAX = 1000.0


TRANSFORMS = [TRANSFORM_WP, TRANSFORM_DS]

DEFAULT_SECTIONS = {
    TRANSFORM_DS: ("matter_galaxy_power", "ds"),
    TRANSFORM_WP: (     "galaxy_power",   "wp"),
}

OUTPUT_NAMES = {
    TRANSFORM_WP:  "bin_{}",
    TRANSFORM_DS:  "bin_{}",
}


# Bias q and order mu parameters for transform
_TRANSFORM_PARAMETERS = {
    TRANSFORM_WP: (0.0, 0.0),
    TRANSFORM_DS: (0.0, 2.0),
}


class LogInterp:
    """
    This is a helper object that interpolates into f(x) where x>0.
    If all f>0 then it interpolates log(f) vs log(x).  If they are all f<0 then it
    interpolate log(-f) vs log(x).  If f is mixed or has some f=0 then it just interpolates
    f vs log(x).

    """

    def __init__(self, angle, spec, kind):
        if np.all(spec > 0):
            self.interp_func = scipy.interpolate.interp1d(
                np.log(angle), np.log(spec), kind, bounds_error=False, fill_value='extrapolate')
            self.interp_type = 'loglog'
        elif np.all(spec < 0):
            self.interp_func = scipy.interpolate.interp1d(
                np.log(angle), np.log(-spec), kind, bounds_error=False, fill_value='extrapolate')
            self.interp_type = 'minus_loglog'
        else:
            self.interp_func = scipy.interpolate.interp1d(
                np.log(angle), spec, kind, bounds_error=False, fill_value='extrapolate')
            self.interp_type = 'log_ang'

    def __call__(self, angle):
        if self.interp_type == 'loglog':
            spec = np.exp(self.interp_func(np.log(angle)))
        elif self.interp_type == 'minus_loglog':
            spec = -np.exp(self.interp_func(np.log(angle)))
        else:
            assert self.interp_type == 'log_ang'
            spec = self.interp_func(np.log(angle))
        return spec


class Transformer:
    """
    Class to build Hankel Transformers that convert from 3D power spectra to correlation functions.
    Several transform types are allowed, depending whether you are using cosmic shear, clustering, or
    galaxy-galaxy lensing.
    """

    def __init__(self, transform_type, n, k_min, k_max,
                 rp_min, rp_max, lower=1.0, upper=-2.0):

        # We use a fixed ell grid in log space and will interpolate/extrapolate our inputs onto this
        # grid. We typically use a maximum ell very much higher than the range we have physical values
        # for.  The exact values there do not matter, but they must be not have a sharp cut-off to avoid
        # oscillations at small angle.
        self.k_min = k_min
        self.k_max = k_max
        k = np.logspace(np.log10(k_min), np.log10(k_max), n)
        self.k = k
        dlogr = np.log(k[1]) - np.log(k[0])

        # pyfftlog has several options about how the theta and ell values used are chosen.
        # This option tells it to pick them to minimize ringing.
        kropt = 1

        # The parameters of the Hankel transform depend on the type.
        # They are defined in a dict at the top of the file
        self.q, self.mu = _TRANSFORM_PARAMETERS[transform_type]

        # Prepare the Hankel transform.
        self.kr, self.xsave = pyfftlog.fhti(
            n, self.mu, dlogr, q=self.q, kropt=kropt)

        # We always to the inverse transform, from Fourier->Real.
        self.direction = -1

        # Some more fixed values.
        self.rp_min = rp_min
        self.rp_max = rp_max
        self.lower = lower
        self.upper = upper

        # work out the effective rp values.
        nc = 0.5 * (n + 1)
        log_kmin = np.log(k_min)
        log_kmax = np.log(k_max)
        log_kmid = 0.5 * (log_kmin + log_kmax)
        k_mid = np.exp(log_kmid)
        r_mid = self.kr / k_mid
        x = np.arange(n)

        # And the effective separations of the output
        self.rp = np.exp((x - nc) * dlogr) * r_mid
        #self.rp = np.degrees(self.rp_rad) * 60.0
        self.range = (self.rp > self.rp_min) & (self.rp < self.rp_max)


    def __call__(self, k_in, pk_in):
        """Convert the input k and P(k) points to the points this transform requires, and then
        transform."""

        # Sample onto self.ell
        pk = self._interpolate_and_extrapolate_pk(k_in, pk_in)

        if self.q == 0:
            xi = pyfftlog.fht(self.k * pk, self.xsave,
                              tdir=self.direction) * (2 * np.pi) / self.rp
        else:
            xi = pyfftlog.fhtq(self.k * pk, self.xsave,
                               tdir=self.direction) * (2 * np.pi) / self.rp

        return self.rp[self.range], xi[self.range]

    def _interpolate_and_extrapolate_pk(self, k, pk):
        """Extrapolate and interpolate the input ell and cl to the default points for this transform"""
        k_min = k[0]
        k_max = k[-1]
        interpolator = LogInterp(k, pk, 'linear')
        pk_out = interpolator(self.k)
        #bad_low = np.isnan(pk_out) & (self.k < k_min)
        #bad_high = np.isnan(pk_out) & (self.k > k_max)

        #pk_out[bad_low] = pk[0] * (self.k[bad_low] / k_min)**self.lower
        #pk_out[bad_high] = pk[-1] * (self.k[bad_high] / k_max)**self.upper

        return pk_out


class CosmosisTransformer(Transformer):
    """
    Subclass of the Transformer object above specialised to cosmosis - gets its configuration
    and input/output from cosmosis data blocks.
    """

    def __init__(self, corr_type, options):
        # The type of transform to perform

        # Where to get/put the input/outputs
        default_input, default_output = DEFAULT_SECTIONS[corr_type]
        self.corr_type = corr_type

        self.input_section = options.get_string(
            option_section, "input_section_name", default_input)

        self.output_section = options.get_string(
            option_section, "output_section_name", default_output)

        if options.has_value(option_section, "suffixes"):
            self.suffixes = np.asarray([options[option_section, "suffixes"]]).flatten()
            self.nbins = len(self.suffixes)
            self.sample = options.get_string(option_section, "sample")
        else:
            self.suffixes = None
            self.nbins = None
            self.sample = options.get_string(option_section, "sample")

        # We don't use the default= keyword above because it breaks
        # for the xip / xim
        if self.input_section == "":
            self.input_section = default_input
        if self.output_section == "":
            self.output_section = default_output

        # Parameters of the transform
        self.n = options.get_int(option_section, "n_transform", DEFAULT_N_TRANSFORM)
        k_min = options.get_double(
            option_section, "k_min_extrapolate", DEFAULT_K_MIN)
        k_max = options.get_double(
            option_section, "k_max_extrapolate", DEFAULT_K_MAX)
        rp_min = options.get_double(
            option_section, "rp_min", DEFAULT_RP_MIN)
        rp_max = options.get_double(
            option_section, "rp_max", DEFAULT_RP_MAX)

        self.output_name = OUTPUT_NAMES[corr_type]

        super().__init__(
            corr_type, self.n, k_min, k_max, rp_min, rp_max)

    def __call__(self, block):

        # Choose the bin values to go up to.  Different modules might specify this in different ways.
        # They might have one nbin value (for cosmic shear and clustering) or two (for GGL)
        if self.corr_type == "ds":
            density = block["density", "mean_density0"]/1e12
        else:
            density = 1.0

        if self.nbins is None:
            nbins = block[self.sample, "nbin"]
        else:
            nbins = self.nbins
        # Loop through bin pairs and see if P(k) exists for all of them
        for i in range(nbins):

            b1 = i + 1

            # The key name for each bin
            output_name = self.output_name.format(b1)

            if self.suffixes is not None:
                input_section = f"{self.input_section}_{self.suffixes[i]}"
            else:
                input_section = self.input_section
            # Read the input k.
            k = block[input_section, "k_h"]
            z = block[input_section, "z"]

            # Read input P(k) from data block.
            pk = block[input_section, "p_k"]

            # Compute the transform.  Calls the earlier __call__ method above.
            xi = np.zeros((z.size, len(self.rp[self.range])))

            for j in range(len(z)):
                rp, xi[j,:] = super().__call__(k, pk[j,:])
            # Integrate over n(z)
            nz = self.load_kernel(block, self.sample, b1, z, 0.0)
            xi = simpson(nz[:,np.newaxis]*xi*density, z, axis=0)
            # Save results back to cosmosis
            block[self.output_section, output_name] = xi

        block[self.output_section, "nbin"] = nbins
        block[self.output_section, "sample"] = self.sample
        block[self.output_section, "rp"] = rp
        block[self.output_section, "sep_name"] = "rp"
        block[self.output_section, "save_name"] = self.corr_type

    def load_kernel(self, block, kernel_section, bin, z_ext, extrapolate_option):

        z_obs = block[kernel_section, "z"]
        obs_in = block[kernel_section, f"bin_{bin}"]
        inter_func = scipy.interpolate.interp1d(z_obs, obs_in, kind="linear", fill_value=extrapolate_option, bounds_error=False)
        kernel_ext = inter_func(z_ext)

        return kernel_ext


def setup(options):
    # xi, gamma, or w - defines what type of transform to do.
    corr_type = options.get_string(option_section, "corr_type")
    if corr_type not in TRANSFORMS:
        raise ValueError("Parameter transform in pk_to_corr must be one of {}".format(
            ", ".join(TRANSFORMS)))

    # The transformer object, which stores all the constants of the transform.
    # Further parameters of the transform are chosen in the __init__ function of CosmosisTransformer
    # in the code above.
    transformer = CosmosisTransformer(corr_type, options)

    return transformer


def execute(block, config):
    transformer = config
    transformer(block)
    return 0



import collections
import matplotlib.pyplot as pl
import numpy as np
import scipy.interpolate
from astropy.units import Quantity
from cosmosis.datablock import names, option_section
from glob import glob


def load_covariance_2d(covfile, nobsbins, nrbins, exclude=None):
    """Load covariance from a 2d matrix-like covariance file

    Parameters
    ----------
    covfile : str
        filename of covariance file
    exclude : list-like, optional
        bins to be excluded from the covariance. Only supports one set
        of bins, which is excluded from the entire dataset

    Returns
    -------
    cov : array, shape (Nobsbins,Nobsbins,Nrbins,Nrbins)
        four-dimensional covariance matrix
    icov : array, shape (Nobsbins,Nobsbins,Nrbins,Nrbins)
        four-dimensional inverse covariance matrix
    likenorm : float
        constant additive term for the likelihood
    esd_err : array, shape (Nobsbins,Nrbins)
        square root of the diagonal elements of the covariance matrix
    cov2d : array, shape (Nobsbins*Nrbins,Nobsbins*Nrbins)
        re-shaped covariance matrix, for plotting or other uses
    """
    cov2d = np.loadtxt(covfile)
    cov2d, nrbins = covariance_excluder(cov2d, nobsbins, nrbins, exclude)
    cov, icov, likenorm, signal_err, cor = covariance_aux_numbers(cov2d, nobsbins, nrbins)

    return cov, icov, likenorm, signal_err, cov2d, cor


def covariance_aux_numbers(cov2d, nobsbins, nrbins):
    try:
        icov_in = np.linalg.inv(cov2d)
    except:
        warnings.warn(
            'Standard matrix inversion failed, using the Moore-Penrose' \
                ' pseudo-inverse',
            RuntimeWarning)
        icov_in = np.linalg.pinv(cov2d)
    cor_in = cov2d/np.sqrt(np.outer(np.diag(cov2d), np.diag(cov2d.T)))
    cov = []
    icov = []
    cor = []
    signal_err = []
    for i in range(nobsbins):
        cov.append([])
        icov.append([])
        cor.append([])
        for j in range(nobsbins):
            indices_i = nrbins[:i].sum()
            indices_j = nrbins[:j].sum()
            s_ = np.s_[indices_i:indices_i+nrbins[i], indices_j:indices_j+nrbins[j]]
            cov[-1].append(cov2d[s_])
            icov[-1].append(icov_in[s_])
            cor[-1].append(cor_in[s_])
            if i==j:
                signal_err.append(np.diag(cov2d[s_])**0.5) # This needs to be in shape (nbins,)
    # product of the determinants
    prod_detC = np.linalg.det(cov2d)
    try:
        posdet = np.linalg.cholesky(cov2d)
    except np.linalg.LinAlgError:
        print('WARNING: Covariance matrix is not positive definite!')
    # likelihood normalization
    likenorm = -(nobsbins**2*np.log(2*np.pi) + np.log(prod_detC)) / 2
    return cov, icov, likenorm, signal_err, cor


def covariance_excluder(cov2d, nobsbins, nrbins, exclude):
    if exclude is None:
        return cov2d, nrbins
    if not hasattr(exclude, '__iter__'):
        exclude = [exclude]
    nexcl = len(exclude)
    idx = np.array([[nrbins[:a].sum()+b for b in exclude] for a in range(nobsbins)])
    cov2d = np.delete(cov2d, idx, axis=0)
    cov2d = np.delete(cov2d, idx, axis=1)
    for i,n in enumerate(nrbins):
        if len(exclude[exclude >= n]) != 0:
            nrbins[i] = n - len(exclude[exclude < n])
        else:
            nrbins[i] = n - nexcl
    return cov2d, nrbins


def load_datapoints_2d(datafiles, datacols, exclude=None):

    datafiles = sorted(glob(datafiles))
    nobsbins = len(datafiles)
    x = np.empty(nobsbins, dtype=object)
    y = np.empty(nobsbins, dtype=object)
    for i,file in enumerate(datafiles):
        x[i] = np.loadtxt(file, usecols=datacols[0])
        y[i] = np.loadtxt(file, usecols=datacols[1])

    if len(datacols) == 3:
        oneplusk = np.empty(nobsbins, dtype=object)
        for i,file in enumerat(datafiles):
            oneplusk[i] = np.loadtxt(file, usecols=[datacols[2]])
        y /= oneplusk
    if exclude is not None:
        x = np.array([np.array([Ri[j] for j in range(len(Ri))
                      if j not in exclude]) for Ri in x], dtype=object)
        y = np.array([np.array([esdi[j] for j in range(len(esdi))
                        if j not in exclude]) for esdi in y], dtype=object)
    return x, y, nobsbins


def load_data(data_dict):

    # first load without excluding anything to define Nrbins, Nobsbins
    x, y, nobsbins = load_datapoints_2d(data_dict['data'], data_dict['columns'])
    nrbins = np.array([sh.size for sh in y])
    cov = load_covariance_2d(
        data_dict['covariance'],nobsbins, nrbins, data_dict['exclude'])
    # Now exclude data points
    x, y, nobsbins = load_datapoints_2d(
        data_dict['data'], data_dict['columns'], data_dict['exclude'])
    nrbins = np.array([sh.size for sh in y])

    return x, y, cov, nobsbins, nrbins


def get_theory_point(x, y, mode='interpolate', interpolated_x=None, bin_edges=None, weighting=None):
    intp = scipy.interpolate.interp1d(np.log(x), y, kind='linear')
    if mode == 'interpolate':
        result = intp(np.log(interpolated_x))
    elif mode == 'integrate':
        result = np.zeros(bin_edges.size-1)
        for i in range(bin_edges.size-1):
            mask = (bin_edges[i] <= x) & (x < bin_edges[i+1])
            if isinstance(weighting, np.ndarray):
                w = weighting[mask]
                norm = np.trapz(w, x[mask])
            else:
                w = 1.0
                norm = 1.0
            result[i] = np.trapz(w*y[mask], x[mask])/norm
    else:
        raise ValueError(f'Mode {mode} not supported.')
    return result




def setup(options):

    # TO-DO: - Implement exclude option
    #        - Implement scaling of observables to the cosmology of the data
    #        -

    input_sections = options.get_string(option_section, 'input_section_names').split()

    data_dict = {}
    data_dict['data'] = options[option_section, 'data']
    data_dict['covariance'] = options[option_section, 'covariance']
    data_dict['columns'] = np.asarray([options[option_section, 'columns']]).flatten()
    #data_dict['exclude'] = np.asarray([options[option_section, 'exclude']]).flatten() # Maybe do with classic scale-cuts as well?
    data_dict['exclude'] = None#options.get_bool(option_section, 'exclude', default=None)# Maybe do with classic scale-cuts as well?
    unit = options[option_section, 'unit']

    x, data_vectors, cov, nobsbins, nrbins = load_data(data_dict)
    cov, inv_cov, likenorm, err, cov2d, cor = cov

    binning_mode = options.get_string(option_section, 'binning_mode', default='interpolate')
    if binning_mode == 'integrate':
        edges = np.array([np.empty(xi.size+1) for xi in x], dtype=object)
        for i,_ in enumerate(edges):
            xi = x[i].astype(float)
            edges[i][1:-1] = (np.log10(xi[1:]) + np.log10(xi[:-1])) * 0.5
            # Compute the first and last by making them symmetric
            edges[i][0] = 2.0 * np.log10(xi[0]) - edges[i][1]
            edges[i][-1] = 2.0 * np.log10(xi[-1]) - edges[i][-2]
        edges = 10.0**edges
    else:
        edges = x.copy()

    like_name = options.get_string(option_section, 'like_name')
    keep_theory_vector = options.get_bool(option_section, 'keep_theory_vector', False)
    return input_sections, cov, inv_cov, x, data_vectors, nobsbins, nrbins, binning_mode, edges, like_name, keep_theory_vector, unit


def execute(block, config):
    input_sections, cov, inv_cov, x, data_vectors, nobsbins, nrbins, binning_mode, edges, like_name, keep_theory_vector, unit = config

    nbins = 0
    for input in input_sections:
        nbins += block[input, 'nbin']
    if nbins != nobsbins:
        raise ValueError('Number of bins in data vector is not the same as the total number of theory bins!')

    count = 0
    theory_vectors = np.empty(nobsbins, dtype=object)
    for i,input in enumerate(input_sections):
        nbins = block[input, 'nbin']
        for j in range(nbins):
            if input in ['wp', 'ds']:
                data_xi = x[count].astype(float)
                data_xi = Quantity(data_xi, unit=unit)
                data_x = data_xi.to('Mpc').value
                theory_x_in = block[input, 'rp']
            else:
                data_x = x[count].astype(float)
                theory_x_in = block[input, f'obs_{j+1}'] * block['cosmological_parameters', 'h0']
            theory_y_in = block[input, f'bin_{j+1}']

            theory_vectors[count] = get_theory_point(theory_x_in, theory_y_in,
                                            mode=binning_mode,
                                            interpolated_x=data_x,
                                            bin_edges=edges[count])
            #pl.plot(data_x, data_vectors[count])
            #pl.plot(data_x, theory_vectors[count])
            #pl.plot(data_x, data_vectors[count]/theory_vectors[count])
            count += 1

        #pl.xscale('log')
        #pl.yscale('log')
        #pl.show()
        #pl.clf

    residuals = data_vectors - theory_vectors
    chi2 = np.array([np.dot(residuals[m], np.dot(inv_cov[m][n], residuals[n]))
                    for m in range(nobsbins) for n in range(nobsbins)]).sum()
    like = -0.5*chi2

    block[names.data_vector, like_name+'_CHI2'] = chi2
    block[names.likelihoods, like_name+'_LIKE'] = like

    if keep_theory_vector:
        block[names.data_vector, 'theory'] = np.hstack(theory_vectors.flatten())
        block[names.data_vector, 'data'] = np.hstack(data_vectors.flatten())
        #block[names.data_vector, like_name + "_inverse_covariance"] = inv_cov.flatten()

    return 0

def cleanup(config):
    pass



import numpy as np
from cosmosis.datablock import names, option_section
from cosmosis.datablock.cosmosis_py import errors


def setup(options):
    like_name = options.get_string(option_section, "like_name")
    input_section_name = options.get_string(option_section, "input_section_name", default="likelihood")

    return like_name, input_section_name

def execute(block, config):
    like_name, input_section_name = config

    d = block[input_section_name, "data"]
    mu = block[input_section_name, "theory"]
    #print('data:',np.round(d,decimals=4))
    #print('theory:',np.round(mu,decimals=4))
    # dir(block)
    # for key in block.keys('hod_parameters_bright'):
    #     print(key, '=', block[key])

    inv_cov = block[input_section_name, "inv_covariance"]
    r = d - mu

    chi2 = float(r @ inv_cov @ r)
    ln_like = -0.5*chi2

    block[names.data_vector, like_name+"_CHI2"] = chi2
    block[names.likelihoods, like_name+"_LIKE"] = ln_like

    return 0

def clean(config):
    pass



import pytest
import numpy as np
from unittest.mock import MagicMock
from onepower import (
    HaloOccupationDistribution,
    Cacciato,
    Simple,
    Zehavi,
    Zheng,
    Zhai,
    load_data,
)


@pytest.fixture
def setup_data():
    rng = np.random.default_rng(seed=42)
    mass = np.logspace(12, 15, 100)
    dndlnm = rng.random((15, 100))
    halo_bias = rng.random((15, 100))
    z_vec = np.linspace(0, 3, 15)
    cosmo = MagicMock()
    cosmo.h = 0.7
    hod_settings = {
        'observables_file': None,
        'obs_min': np.atleast_1d(8.0),
        'obs_max': np.atleast_1d(12.0),
        'zmin': np.atleast_1d(0.0),
        'zmax': np.atleast_1d(0.2),
        'nz': 15,
        'nobs': 300,
        'observable_h_unit': '1/h^2',
    }
    return mass, dndlnm, halo_bias, z_vec, cosmo, hod_settings


def test_load_data(datadir):
    file_name = f'{datadir}/test_data.txt'
    data = np.array([[0.1, 8.0, 12.0], [0.2, 8.5, 12.5]])
    np.savetxt(file_name, data)
    z_data, obs_min, obs_max = load_data(file_name)
    assert np.allclose(z_data, data[:, 0])
    assert np.allclose(obs_min, data[:, 1])
    assert np.allclose(obs_max, data[:, 2])


def test_hod_initialization_and_quantities(setup_data):
    mass, dndlnm, halo_bias, z_vec, cosmo, hod_settings = setup_data
    hod = HaloOccupationDistribution(
        cosmo=cosmo,
        mass=mass,
        dndlnm=dndlnm,
        halo_bias=halo_bias,
        z_vec=z_vec,
        hod_settings=hod_settings,
    )

    assert hod.cosmo == cosmo
    assert hod.mass.shape == (1, 1, len(mass))
    assert np.allclose(hod.z_vec, z_vec)
    assert hod.dndlnm.shape == dndlnm.shape
    assert hod.halo_bias.shape == halo_bias.shape

    with pytest.raises(NotImplementedError):
        hod._compute_hod_cen
    with pytest.raises(NotImplementedError):
        hod._compute_hod_sat

    obs = hod.obs
    assert obs.shape == (hod.nbins, hod.nz, 1, hod.nobs)

    dndlnm_int = hod.dndlnm_int
    assert dndlnm_int.shape == (1, hod.nz, len(mass))

    halo_bias_int = hod.halo_bias_int
    assert halo_bias_int.shape == (1, hod.nz, len(mass))

    assert hod.data is None
    assert hod.nobs == hod_settings['nobs']
    assert hod.nbins == len(hod_settings['obs_min'])
    assert hod.nz == hod_settings['nz']

    z = hod.z
    assert z.shape == (hod.nbins, hod.nz)

    log_obs_min = hod.log_obs_min
    assert log_obs_min.shape == (hod.nbins, hod.nz)

    log_obs_max = hod.log_obs_max
    assert log_obs_max.shape == (hod.nbins, hod.nz)

    rng = np.random.default_rng(seed=42)
    data = rng.random((hod.nbins, hod.nz))
    interpolated_data = hod._interpolate(data)
    assert interpolated_data.shape == (hod.nbins, len(z_vec))


def test_hod_quantities_with_file(setup_data, datadir):
    mass, dndlnm, halo_bias, z_vec, cosmo, hod_settings = setup_data
    file_name = f'{datadir}/test_data.txt'
    data = np.array([[0.1, 8.0, 12.0], [0.2, 8.5, 12.5]])
    np.savetxt(file_name, data)
    hod_settings['observables_file'] = file_name

    hod = HaloOccupationDistribution(
        cosmo=cosmo,
        mass=mass,
        dndlnm=dndlnm,
        halo_bias=halo_bias,
        z_vec=z_vec,
        hod_settings=hod_settings,
    )

    obs = hod.obs
    assert obs.shape == (hod.nbins, hod.nz, 1, hod.nobs)

    dndlnm_int = hod.dndlnm_int
    assert dndlnm_int.shape == (1, hod.nz, len(mass))

    halo_bias_int = hod.halo_bias_int
    assert halo_bias_int.shape == (1, hod.nz, len(mass))

    assert hod.nobs == hod_settings['nobs']
    assert hod.nbins == len(hod_settings['obs_min'])
    assert hod.nz == len(data)

    z = hod.z
    assert z.shape == (hod.nbins, hod.nz)

    log_obs_min = hod.log_obs_min
    assert log_obs_min.shape == (hod.nbins, hod.nz)

    log_obs_max = hod.log_obs_max
    assert log_obs_max.shape == (hod.nbins, hod.nz)

    rng = np.random.default_rng(seed=42)
    data = rng.random((hod.nbins, hod.nz))
    interpolated_data = hod._interpolate(data)
    assert interpolated_data.shape == (hod.nbins, len(z_vec))


def test_cacciato_properties(setup_data):
    mass, dndlnm, halo_bias, z_vec, cosmo, hod_settings = setup_data
    cacciato = Cacciato(
        cosmo=cosmo,
        mass=mass,
        dndlnm=dndlnm,
        halo_bias=halo_bias,
        z_vec=z_vec,
        hod_settings=hod_settings,
        A_cen=0.0,
        A_sat=0.0,
    )

    assert isinstance(cacciato, Cacciato)
    assert cacciato.Obs_norm_c == 10.0**9.95
    assert cacciato.M_char == 10.0**11.24

    assert cacciato.number_density.shape == (cacciato.nbins, cacciato.nz)
    assert cacciato.number_density_cen.shape == (cacciato.nbins, cacciato.nz)
    assert cacciato.number_density_sat.shape == (cacciato.nbins, cacciato.nz)
    assert cacciato.avg_halo_mass.shape == (cacciato.nbins, cacciato.nz)
    assert cacciato.avg_halo_mass_cen.shape == (cacciato.nbins, cacciato.nz)
    assert cacciato.avg_halo_mass_sat.shape == (cacciato.nbins, cacciato.nz)
    assert cacciato.galaxy_linear_bias.shape == (cacciato.nbins, cacciato.nz)
    assert cacciato.galaxy_linear_bias_cen.shape == (cacciato.nbins, cacciato.nz)
    assert cacciato.galaxy_linear_bias_sat.shape == (cacciato.nbins, cacciato.nz)
    assert cacciato.f_c.shape == (cacciato.nbins, cacciato.nz)
    assert cacciato.f_s.shape == (cacciato.nbins, cacciato.nz)
    assert cacciato.hod.shape == (cacciato.nbins, cacciato.nz, mass.size)
    assert cacciato.hod_cen.shape == (cacciato.nbins, cacciato.nz, mass.size)
    assert cacciato.hod_sat.shape == (cacciato.nbins, cacciato.nz, mass.size)
    assert cacciato.stellar_fraction.shape == (cacciato.nbins, cacciato.nz, mass.size)
    assert cacciato.stellar_fraction_cen.shape == (
        cacciato.nbins,
        cacciato.nz,
        mass.size,
    )
    assert cacciato.stellar_fraction_sat.shape == (
        cacciato.nbins,
        cacciato.nz,
        mass.size,
    )
    assert cacciato.obs_func.shape == (cacciato.nbins, cacciato.nz, cacciato.nobs)
    assert cacciato.obs_func_cen.shape == (cacciato.nbins, cacciato.nz, cacciato.nobs)
    assert cacciato.obs_func_sat.shape == (cacciato.nbins, cacciato.nz, cacciato.nobs)


def test_simple_properties(setup_data):
    mass, dndlnm, halo_bias, z_vec, cosmo, hod_settings = setup_data
    simple = Simple(
        cosmo=cosmo,
        mass=mass,
        dndlnm=dndlnm,
        halo_bias=halo_bias,
        z_vec=z_vec,
        hod_settings=hod_settings,
        A_cen=0.0,
        A_sat=0.0,
    )

    assert isinstance(simple, Simple)
    assert simple.log10_Mmin == 12.0
    assert simple.log10_Msat == 13.0
    assert simple.alpha == 1.0
    assert simple._compute_hod_cen.shape == (1, z_vec.size, mass.size)
    assert simple._compute_hod_sat.shape == (1, z_vec.size, mass.size)
    assert np.allclose(
        simple.stellar_fraction, np.zeros((simple.nbins, simple.nz, mass.size))
    )
    assert np.allclose(
        simple.stellar_fraction_cen, np.zeros((simple.nbins, simple.nz, mass.size))
    )
    assert np.allclose(
        simple.stellar_fraction_sat, np.zeros((simple.nbins, simple.nz, mass.size))
    )


def test_zehavi_properties(setup_data):
    mass, dndlnm, halo_bias, z_vec, cosmo, hod_settings = setup_data
    zehavi = Zehavi(
        cosmo=cosmo,
        mass=mass,
        dndlnm=dndlnm,
        halo_bias=halo_bias,
        z_vec=z_vec,
        hod_settings=hod_settings,
        A_cen=0.0,
        A_sat=0.0,
    )

    assert isinstance(zehavi, Zehavi)
    assert zehavi.log10_Mmin == 12.0
    assert zehavi.log10_Msat == 13.0
    assert zehavi.alpha == 1.0
    assert zehavi._compute_hod_cen.shape == (1, z_vec.size, mass.size)
    assert zehavi._compute_hod_sat.shape == (1, z_vec.size, mass.size)


def test_zheng_properties(setup_data):
    mass, dndlnm, halo_bias, z_vec, cosmo, hod_settings = setup_data
    zheng = Zheng(
        cosmo=cosmo,
        mass=mass,
        dndlnm=dndlnm,
        halo_bias=halo_bias,
        z_vec=z_vec,
        hod_settings=hod_settings,
        A_cen=0.0,
        A_sat=0.0,
    )

    assert isinstance(zheng, Zheng)
    assert zheng.log10_Mmin == 12.0
    assert zheng.log10_M0 == 12.0
    assert zheng.log10_M1 == 13.0
    assert zheng.alpha == 1.0
    assert zheng.sigma == 0.15
    assert zheng._compute_hod_cen.shape == (1, z_vec.size, mass.size)
    assert zheng._compute_hod_sat.shape == (1, z_vec.size, mass.size)


def test_zhai_properties(setup_data):
    mass, dndlnm, halo_bias, z_vec, cosmo, hod_settings = setup_data
    zhai = Zhai(
        cosmo=cosmo,
        mass=mass,
        dndlnm=dndlnm,
        halo_bias=halo_bias,
        z_vec=z_vec,
        hod_settings=hod_settings,
        A_cen=0.0,
        A_sat=0.0,
    )

    assert isinstance(zhai, Zhai)
    assert zhai.log10_Mmin == 13.68
    assert zhai.log10_Msat == 14.87
    assert zhai.log10_Mcut == 12.32
    assert zhai.alpha == 0.41
    assert zhai.sigma == 0.82
    assert zhai._compute_hod_cen.shape == (1, z_vec.size, mass.size)
    assert zhai._compute_hod_sat.shape == (1, z_vec.size, mass.size)
